/*
 * Copyright (c) 2004, Bull S.A..  All rights reserved.
 * Created by: Sebastien Decugis

 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc., 59
 * Temple Place - Suite 330, Boston MA 02111-1307, USA.

 * This sample aims to demonstrate a bug in pthread_cond_timedwait
 
*/
 
 /* We are testing conformance to IEEE Std 1003.1, 2003 Edition */
 #define _POSIX_C_SOURCE 200112L

 #include <pthread.h>
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h> 
 #include <unistd.h>
 #include <string.h>

 #include <semaphore.h>
 #include <errno.h>
 #include <time.h> /* we need the clock_settime routine */
 
typedef struct
{
	pthread_mutex_t * pmtx;
	pthread_cond_t * pcnd;
	struct timespec * timeout;
	int ctrl;
	int rc;
} datatest_t;


#define FLAG_TIMEDOUT  (1 << 15) /* Will be set if the function exited with a timeout error */
#define FLAG_AWAKEN    (1 << 14) /* Will be set if the function exited with no error and the next flag was set */
#define FLAG_CONDWAKE  (1 << 13) /* the boolean associated with the condition */
#define FLAG_INWAIT    (1 << 12) /* Child thread got the mutex locked */


#define UNRESOLVED(rc, str) \
{ \
	printf("UNRESOLVED CASE:\n Error %d - %s\n Message %s\n", rc, strerror(rc), str);  \
	exit(1); \
}

 
 /****
 * test_timeout
 *
 */
void * test_timeout(void * arg)
{
	int ret;
	datatest_t * dt = (datatest_t *)arg;
	
	/* lock the test mutex */
	ret = pthread_mutex_lock(dt->pmtx);
	if (ret != 0)
	{  UNRESOLVED(ret, "Unable to lock mutex in test_timeout");  }
	
	/* Signal the parent thread that we are ready to wait */
	dt->ctrl |= FLAG_INWAIT;
	
	/* Enter the timed wait */
	do
	{
		dt->rc = pthread_cond_timedwait(dt->pcnd, dt->pmtx, dt->timeout);
	} while ((dt->rc == 0) && ((dt->ctrl & FLAG_CONDWAKE) == 0));
	
	/* Set the flags */
	if (dt->rc == 0)
	{
		dt->ctrl |= FLAG_AWAKEN;
	}
	if (dt->rc == ETIMEDOUT)
	{
		dt->ctrl |= FLAG_TIMEDOUT;
	}
	
	/* Unlock the mutex and exit */
	ret = pthread_mutex_unlock(dt->pmtx);
	if (ret != 0)
	{  UNRESOLVED(ret, "Unable to unlock the test mutex in test_timeout");  }
	
	return NULL;
}

datatest_t DT;

/****
 * do_cs_test
 *  This function will take care of the clock testing.
 */
int do_cs_test(void)
{
	int ret;
	int result=0;
	
	pthread_condattr_t ca;
	pthread_mutex_t mtx;
	pthread_cond_t cnd;
	
	pthread_t th;
	
	struct timespec ts, timeout;
	
	clockid_t  cid;

	/* We are going to initialize the cond vars and the mutexes */
	DT.pmtx = &mtx;
	DT.pcnd = &cnd;
	ret = pthread_mutex_init(&mtx, NULL);
	if (ret != 0)
	{  UNRESOLVED(ret, "Unable to initialize a default mutex");  }

	ret = pthread_condattr_init(&ca);
	if (ret != 0)
	{  UNRESOLVED(ret, "Unable to initialize cond attribute object");  }

	ret = pthread_cond_init(&cnd, &ca);
	if (ret != 0)
	{  UNRESOLVED(ret, "Unable to initialize the default attribute conditional variable.");  }

	
	DT.ctrl = 0;
	DT.rc   = 0;
	
	printf("Data initialized successfully for CS test.\n");
	
	/* Get the default clock ID */
	ret = pthread_condattr_getclock(&ca, &cid);
	if (ret != 0)
	{  UNRESOLVED(ret, "Unable to get clockid from the default cond attribute object");  }
	
	/* Check whether we can set the system clock */
	ret = clock_gettime(cid, &ts);
	if (ret != 0)
	{  UNRESOLVED(errno, "Unable to get clock time");  }
	
	ret = clock_settime(cid, &ts);
	if (ret != 0)
	{
		UNRESOLVED(errno, "Unable to set clock - please run as root\n Warning: the system time will be corrupted\n");
	}
	else /* We can do the test */
	{
		printf("clock_settime succeeded\n");
		
		/* Prepare the timeout parameters */
		timeout.tv_nsec = 0;
		timeout.tv_sec = ts.tv_sec + 20; /* 20 sec timeout */
		DT.timeout = &timeout;
		
		/* create the thread */
		ret = pthread_create(&th, NULL, test_timeout, &DT);
		if (ret != 0)
		{  UNRESOLVED(ret, "Unable to create a thread");  }
		
		/* Lock the mutex and make sure the thread is in wait */
		ret = pthread_mutex_lock(&mtx);
		if (ret != 0)  {  UNRESOLVED(ret, "Unable to lock a mutex");  }
		while ((DT.ctrl & FLAG_INWAIT) == 0)
		{
			ret = pthread_mutex_unlock(&mtx);
			if (ret != 0)  {  UNRESOLVED(ret, "Unable to unlock a mutex");  }
			sched_yield();
			ret = pthread_mutex_lock(&mtx);
			if (ret != 0)  {  UNRESOLVED(ret, "Unable to lock a mutex");  }
		}
		
		printf("Thread is created and waiting.\n");
		printf("About to change the default clock value...\n");
		
		/* We re-read the default clock time, to introduce minimal error on the clock */
		ret = clock_gettime(cid, &ts);
		if (ret != 0)  {  UNRESOLVED(errno, "Unable to get clock time");  }
		
		ts.tv_sec += 604800;  /* Exactly 1 week forth */
		
		/* We set the clock to a date after the timeout parameter */
		ret = clock_settime(cid, &ts);
		if (ret != 0)  {  UNRESOLVED(errno, "Unable to set clock time (again)");  }
		
		/* unlock the two mutex, so the thread can exit pthread_cond_timedwait */
		ret = pthread_mutex_unlock(&mtx);
		if (ret != 0)  {  UNRESOLVED(ret, "Unable to unlock a mutex");  }
		
		/* Give the other thread an opportunity to run */
		sched_yield();
		
		printf("Checking that the thread has timed out...\n");
		
		/* Relock the mutexs */
		ret = pthread_mutex_lock(&mtx);
		if (ret != 0)  {  UNRESOLVED(ret, "Unable to lock a mutex");  }
		
		if ((DT.ctrl & FLAG_TIMEDOUT) != 0)
		{
			printf("The cond var timed out (good!)\n");
			/* Join the thread */
			ret = pthread_join(th, NULL);
			if (ret != 0)  {  UNRESOLVED(ret, "Join thread failed");  }
			/* Unlock the mutex */
			ret = pthread_mutex_unlock(&mtx);
			if (ret != 0)  {  UNRESOLVED(ret, "Unable to unlock a mutex");  }
		}
		else
		{
			printf("The cond var did not timed out immediatly (as it should)\n");

			/* Signal the condition */
			DT.ctrl |= FLAG_CONDWAKE;
#ifdef DOSIGNAL			
			ret = pthread_cond_signal(&cnd);
			if (ret != 0)  {  UNRESOLVED(ret, "Unable to signal the condition");  }
#endif
			/* Unlock the mutex and join the thread */
			ret = pthread_mutex_unlock(&mtx);
			if (ret != 0)  {  UNRESOLVED(ret, "Unable to unlock a mutex");  }
			/* Join the thread */
			ret = pthread_join(th, NULL);
			if (ret != 0)  {  UNRESOLVED(ret, "Join thread failed");  }
		}
		
		/* Set the system time back to its value */
		ret = clock_gettime(cid, &ts);
		if (ret != 0)  {  UNRESOLVED(errno, "Unable to get clock time");  }
		
		ts.tv_sec -= 604800;  /* Exactly 1 week back */ 
		
		ret = clock_settime(cid, &ts);
		if (ret != 0)  {  UNRESOLVED(errno, "Unable to set clock time (3rd time)");  }

		printf("Default clock was set back\n");

		/* Results */
		printf("Cond var timedwait return value: %d\n", DT.rc);
		printf("Cond var exited with a timeout : %s\n", (DT.ctrl & FLAG_TIMEDOUT)?"yes":"no");
		printf("Cond var exited when signaled  : %s\n", (DT.ctrl & FLAG_AWAKEN)?"yes":"no");
		
		
	}
	
	/* We can cleanup things now */
	ret = pthread_cond_destroy(&cnd);
	if (ret != 0)
	{  UNRESOLVED(ret, "Cond destroy failed");  }

	ret = pthread_condattr_destroy(&ca);
	if (ret != 0)
	{  UNRESOLVED(ret, "Cond attr destroy failed");  }

	ret = pthread_mutex_destroy(&mtx);
	if (ret != 0)
	{  UNRESOLVED(ret, "Mutex destroy failed");  }
	
	return result;
}

/****
 * Main function
 */
int main(int argc, char * argv[])
{
	long opt_TMR, opt_CS;
	int ret=0;


	printf("Test starting...\n");

	
	opt_TMR=sysconf(_SC_TIMERS);
	opt_CS =sysconf(_SC_CLOCK_SELECTION);
	
	printf("Timers option : %li\n", opt_TMR);
	printf("Clock Selection option : %li\n", opt_CS);
	
	if ((opt_TMR != -1L) && (opt_CS != -1L))
	{
		printf("Starting clock test\n");
		ret = do_cs_test();
	}
	else
	{
		printf("Unable to test on this implementation");
		ret = 1;
	}		
	
	return ret;
}
